#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <curl/curl.h>

#include "arg.h"
#include "repo.h"

void usage(char *);
int download(char *, FILE *);
int checkpack(char *, char *);
int cachepack(char *, char *, struct packs *, int);

int verbose = 0;
int fflag, lflag, sflag, uflag;

char *sickexec[] = { "sick", NULL };

void
usage(char *name)
{
	fprintf(stderr, "usage: %s [-c FILE] [-flsu] [-r URL] [PACK..]\n", name);
	exit(1);
}

struct pack *
addpack(struct packs *plist, char *name, char *version, char *url)
{
	struct pack *p = NULL;

	p = malloc(sizeof(struct pack));
	if (!p) {
		perror("malloc");
		return NULL;
	}

	strncpy(p->url, url, sizeof(p->url));
	strncpy(p->name, name, sizeof(p->name));
	strncpy(p->version, version, sizeof(p->version));

	TAILQ_INSERT_TAIL(plist, p, entries);

	return p;
}

struct repo *
addrepo(struct repos *rlist, char *url)
{
	struct repo *r = NULL;

	r = malloc(sizeof(struct repo));
	if (!r) {
		perror("malloc");
		exit(1);
	}

	r->url = strdup(url);
	if (!r->url) {
		perror("malloc");
		exit(1);
	}

	TAILQ_INSERT_TAIL(rlist, r, entries);

	return r;
}

int
repolist(struct packs *plist, char *local)
{
	int r = 0;
	char fn[PATH_MAX] = "";
	char buf[LINE_MAX] = "";

	char n[LINE_MAX], v[LINE_MAX], u[PATH_MAX];
	FILE *list;

	snprintf(fn, PATH_MAX, "%s/%s", local, DEFLISTFILE);
	list = fopen(fn, "r");
	if (!list) {
		perror(fn);
		exit(1);
	}

	while (fgets(buf, LINE_MAX, list)) {
		r = sscanf(buf, "%s\t%s\t%s", n, v, u);
		if (r < 3) {
			fprintf(stderr, "%s: Invalid format detected\n", fn);
			exit(1);
		}
		addpack(plist, n, v, u);
	}
	fclose(list);
	return 0;
}

int
download(char *url, FILE *fd)
{
	ssize_t len;
	char *fn, *base, *encfile, *encurl;
	CURL *c;
	CURLcode r;

	c = curl_easy_init();
	if (!c)
		return -1;


	fn = strdup(basename(url));
	base = strdup(dirname(url));
	encfile = curl_easy_escape(c, fn, strlen(fn));

	len = strlen(base) + strlen(encfile) + 1;
	encurl = malloc(len + 1);
	if (!encurl) {
		perror("malloc");
		exit(1);
	}
	snprintf(encurl, len + 1, "%s/%s", base, encfile);

	curl_easy_setopt(c, CURLOPT_URL, encurl);
	curl_easy_setopt(c, CURLOPT_WRITEDATA, fd);
	if (verbose)
		fprintf(stderr, "FETCH %s/%s\n", url, fn);

	r = curl_easy_perform(c);
	if (r != CURLE_OK) {
		fprintf(stderr, "%s: %s\n", encurl, curl_easy_strerror(r));
		exit(1);
	}

	free(fn);
	curl_free(encfile);

	return 0;
}

int
checkpack(char fn[PATH_MAX], char *url)
{
	FILE *f;
	int fd[2], out, status;
	pipe(fd);
	if (!fork()) {
                close(0);
                close(1);
                close(fd[1]);
                dup2(fd[0], 0);

	        if ((out = open(fn, O_WRONLY|O_CREAT|O_TRUNC, 0644)) < 0) {
	                perror(fn);
	                return -1;
	        }
		dup2(out, 1);
                execvp(sickexec[0], sickexec);
		perror(sickexec[0]);
	}

	close(fd[0]);
	f = fdopen(fd[1], "w");
	if (!f) {
		perror("pipe");
		exit(1);
	}

	download(url, f);
	fflush(f);
	fclose(f);

	wait(&status);
	if (status) {
		fprintf(stderr, "%s: Pack verification failed\n", basename(fn));
		unlink(fn);
		return -1;
	}
	return 0;
}


int
cachepack(char *name, char *localrepo, struct packs *plist, int untrust)
{
	int ret = 0;
	FILE *f;
	char fn[PATH_MAX];;
	struct pack *p = NULL;
	struct stat sb;

	TAILQ_FOREACH(p, plist, entries) {
		if (!strncmp(p->name, name, PATH_MAX)) {
			snprintf(fn, PATH_MAX, "%s/%s", localrepo, basename(p->url));
			if (!stat(fn, &sb) && !fflag) {
				puts(fn);
				continue;
			}

			if (untrust) {
				f = fopen(fn, "w");
				if (!f) {
					perror(fn);
					exit(1);
				}
				download(p->url, f);
				fflush(f);
				fclose(f);
			} else {
				if (checkpack(fn, p->url)) {
					ret++;
					continue;
				}
			}

			puts(fn);
			break;
		}
	}
	return ret;
}

int
main (int argc, char *argv[])
{
	char *argv0, *n;
	char  cfgfile[PATH_MAX] = DEFCFGFILE;
	char  localrepo[PATH_MAX] = DEFLOCALREPO;
	char  fn[PATH_MAX], url[PATH_MAX];
	FILE *fd;
	struct stat sb;
	struct packs plist;
	struct repos rlist;
	struct pack *p = NULL;
	struct repo *r = NULL;

	TAILQ_INIT(&plist);
	TAILQ_INIT(&rlist);

	fflag = lflag = sflag = 0;

	ARGBEGIN{
	case 'c':
		snprintf(cfgfile, PATH_MAX, "%s", EARGF(usage(argv0)));
		break;
	case 'f':
		fflag = 1;
		break;
	case 'r':
		addrepo(&rlist, EARGF(usage(argv0)));
		break;
	case 's':
		sflag = 1;
		break;
	case 'u':
		uflag = 1;
		break;
	case 'l':
		lflag = 1;
		break;
	case 'v':
		verbose++;
		break;
	default:
		usage(argv0);
	}ARGEND;

	if (!stat(cfgfile, &sb))
		parseconf(&rlist, localrepo, &uflag, cfgfile);

	if (sflag) {
		snprintf(fn, PATH_MAX, "%s/%s", localrepo, DEFLISTFILE);
		fd = fopen(fn, "w");
		if (!fd) {
			perror(fn);
			exit(1);
		}
		TAILQ_FOREACH(r, &rlist, entries) {
			if (verbose)
				fprintf(stderr, "SYNC %s\n", r->url);
			snprintf(url, PATH_MAX, "%s/%s", r->url, DEFLISTFILE);
			download(url, fd);
		}
		fclose(fd);
	}

	repolist(&plist, localrepo);

	if (lflag) {
		TAILQ_FOREACH(p, &plist, entries) {
			printf("%s\t%s\t%s\n", p->name, p->version, verbose ? p->url : "\b\0");
		}
		return 0;
	}

	while ((n = *(argv++)))
		cachepack(n, localrepo, &plist, uflag);

	return 0;
}
